{-# LANGUAGE OverloadedStrings #-}
module CFG.C
       ( instrsToC
       , compileToC
       , generateDecls
       , generateDecl
       , compileToCAll
       , declareExtern
       , preIdleHook
       , DecisionM
       ) where

import           Control.Comonad
import           Data.List
import           Data.Maybe
import           Data.Text (Text)
import qualified Data.Text as T
import           Data.IntSet (IntSet)
import qualified Data.IntSet as S
import           Data.Semigroup
import           Control.Monad.State

import           CFG.IR
import           CFG.Types

-- | Compile multiple functions to C, inckluding declarations
compileToCAll :: Maybe IntSet -- ^ All possible interrupt numbers
              -> [Function [Instr]]
              -> Text
compileToCAll interrupts funs = flip evalState 0 $ do
  funs' <- T.unlines <$> mapM (compileToC interrupts) funs
  idleHook <- preIdleHook interrupts

  return $
    include
    <> declareExtern funs
    <> "\n"
    <> generateDecls funs
    <> "\n"
    <> funs'
    <> idleHook

type DecisionM = State Int

include :: Text
include = "#include <mockup_prelude.cc>\n\n"

compileToC :: Maybe IntSet -> Function [Instr] -> DecisionM Text
compileToC interrupts fun = do
  instrs <- instrsToC interrupts (funKind fun) $ extract fun
  let header KindSubtask = "TASK(" <> funName fun <> ")"
      header KindFunction = "void " <> funName fun <> "()"
      header KindISR = "ISR2(" <> funName fun <> ")"

  return $
       header (funKind fun) <> " {\n"
    <> indent instrs
    <> "}\n"

-- | Declare multiple functions
generateDecls :: [Function a] -> Text
generateDecls = mconcat . map generateDecl

-- | Declare a function
generateDecl :: Function a -> Text
generateDecl fun = case funKind fun of
  KindSubtask  -> "DeclareTask(" <> funName fun <> ");\n"
  KindFunction -> "void " <> funName fun <> "();\n"
  KindISR      -> ""

-- | Declare all used but undefined symbols as extern
declareExtern :: [Function [Instr]] -> Text
declareExtern funs = mconcat $
     [ "extern \"C\" {\n" ]
  ++ map export (calledNames \\ definedNames)
  ++ ["}"]

  where
    export name = "extern void " <> name <> "();\n"
    definedNames = nub $ map funName funs
    calledNames = nub $ concatMap getCalls funs
    getCalls fun = mapMaybe callName $ funAnnotation fun
    callName (Call name _ _) = Just name
    callName _               = Nothing

preIdleHook :: Maybe IntSet -> DecisionM Text
preIdleHook interrupts = do
  triggerCode <- triggerInterrupts interrupts

  return $
       "void PreIdleHook() {\n"
    <> "kout << \"IPs: \" << arch::OS_Idle_ip << ',' << arch::OS_ISR1_ip << ',' << arch::OS_H1_ip << ',' << arch::OS_H2_ip << ',' << arch::OS_H3_ip << endl;"
    <> "  _print_os_state(\"PreIdleHook\", __func__);\n"
    <> triggerCode
    <> "  ShutdownMachine();\n"
    <> "}\n"

triggerInterrupts :: Maybe IntSet -> DecisionM Text
triggerInterrupts Nothing = return ""
triggerInterrupts (Just ints) = do
  triggerCode <- mconcat <$> mapM triggerInterrupt (S.toList ints)

  return $
       "while (1) {\n"
    <> "  int _triggeredSomeInterrupt = 0;\n"
    <> triggerCode
    <> "  if (_triggeredSomeInterrupt == 0) break;\n"
    <> "}\n"

  where triggerInterrupt int = do
          decision <- decisionNumber
          return $
            "  if (_decisionMaker("<> T.pack (show decision) <> ")) {\n"
            <> "    _print_os_state(\"IRQ" <> T.pack (show int) <> "\", __func__);\n"
            <> "    Machine::trigger_interrupt_from_user(" <> T.pack (show int) <> ");\n"
            <> "    _triggeredSomeInterrupt = 1;\n"
            <> "  }\n"

instrsToC :: Maybe IntSet -> FunctionKind -> [Instr] -> DecisionM Text
instrsToC interrupts kind = fmap ((<> "\n") . mconcat . intersperse "\n") . mapM (instrToC interrupts kind)

instrToC :: Maybe IntSet -> FunctionKind -> Instr -> DecisionM Text
instrToC _ _ (Goto ident) = return $ "goto " <> ident <> ";"
instrToC _ _ (Label ident) = return $ ident <> ":"
instrToC _ _ (IfThanElse left right) = do
  decision <- decisionNumber
  return $
    "if (_decisionMaker(" <> T.pack (show decision) <> "))\n" <>
    "  goto " <> left <> ";\n" <>
    "else\n" <>
    "  goto " <> right <> ";"
instrToC _ _ (Call ident TypeFunction _) = return $ ident <> "();"
instrToC interrupts kind (Call ident TypeSyscall isKickoff)  = do
  ints <-
    if kind /= KindISR && not isKickoff
      then triggerInterrupts interrupts
      else return mempty
  return $
    ints
    <> "_print_os_state(\"" <> ident <> "\", __func__);\n"
    <> ident <> "();"
instrToC _ _ Return = return "return;"

indent :: Text -> Text
indent = T.unlines . map ("  " <>) . T.lines

decisionNumber :: DecisionM Int
decisionNumber = do
  n <- get
  put (n+1)
  return n
