{-# OPTIONS_GHC -fno-warn-name-shadowing #-}
{-# LANGUAGE OverloadedStrings, TupleSections #-}

-- | Compile a CFG to an intermediate representation that is easier to compile
--   directory to C
module CFG.IR
       ( Ident
       , CallType(..)
       , Instr(..)
       , flattenCFG
       , compileToIR
       , compileToIRAll
       ) where

import Data.Array
import Data.Graph
import Data.Text (Text)
import Control.Arrow
import Control.Comonad

import CFG.Types

-- | Identifier, used for labels and function names
type Ident = Text

data CallType = TypeFunction | TypeSyscall
              deriving (Show, Eq)

data Instr = Goto Ident
           | Label Ident
           | IfThanElse Ident Ident
           | Call Ident CallType Bool -- ^ The Bool means isKickoff
           | Return
           deriving (Show, Eq)

-- | compile all functions to IR
--
-- This uses unique decision numbers across all functions
compileToIRAll :: [Function CFG] -> [Function (CFG, [Instr])]
compileToIRAll = map compileToIR

compileToIR :: Function CFG -> Function (CFG, [Instr])
compileToIR = extend (runCokleisli (Cokleisli extract &&& Cokleisli flattenCFG))

-- | Compile a CFG to a list of instructions
flattenCFG :: Function CFG
           -> [Instr]
flattenCFG fun = let cfg = extract fun in
  Goto (vertexToIdent cfg (cfgEntryVertex cfg))
  : concatMap (compileBlock cfg) (vertices $ cfgGraph cfg)


-- | Generate instructions corresponding to one vertex/block
compileBlock :: CFG -> Vertex -> [Instr]
compileBlock cfg vertex =
  Label ident : call ++ [gotoSuccs]

  where ident = vertexToIdent cfg vertex
        call =
          let (blk, _, _) = cfgGetBlock cfg vertex
          in case blkCallInfo blk of
            Just (FunctionCall fun) -> [Call fun TypeFunction (blkIsKickoff blk)]
            Just (SystemCall sys)   -> [Call sys TypeSyscall (blkIsKickoff blk)]
            Nothing                 -> []

        gotoSuccs = case cfgGraph cfg ! vertex of
          [] -> Return -- no successors
          [successor] -> Goto (vertexToIdent cfg successor)
          [left, right] -> IfThanElse
                             (vertexToIdent cfg left)
                             (vertexToIdent cfg right)
          _ -> error "More than two successors in ABB"


vertexToIdent :: CFG -> Vertex -> Ident
vertexToIdent cfg vertex = key
  where (_, key, _) = cfgGetBlock cfg vertex
