{-# LANGUAGE LambdaCase, OverloadedStrings #-}
{-# LANGUAGE ViewPatterns #-}

module Search.Search (driver) where

import Data.Map (Map)
import qualified Data.Map as M
import Data.Set (Set)
import qualified Data.Set as S
import Data.Text (Text)
import qualified Data.Text as T
import qualified Data.Text.IO as T
import Data.Maybe
import Data.Monoid
import Data.IORef
import Control.Monad
import System.Exit
import System.Process
import System.IO
import Data.Foldable

data Metadata = Metadata
  { metaFunc :: Text -- | Function Name
  , metaSyscall :: Text -- | Next Syscall
  }

-- | (Statehash, Metadata)
data Vertex = Vertex
  { vertState :: Text
  , vertMeta :: Metadata
  }

instance Eq Vertex where
  Vertex x meta1 == Vertex y meta2 = (x, metaSyscall meta1) == (y, metaSyscall meta2)

instance Ord Vertex where
  compare (Vertex x meta1) (Vertex y meta2) = compare (x, metaSyscall meta1) (y, metaSyscall meta2)

type Graph = Map Vertex (Set Vertex)

type DecisionNumber = Int
type Decision = (Vertex, DecisionNumber)
data DecisionState = Half [Bool] -- Trace
                   | Full

data State = State
  { graph :: Graph
  , currentVertex :: Maybe Vertex
  , decisions :: Map Decision DecisionState
  , curTrace :: [Bool]
  }

addVertex :: Vertex -> State -> State
addVertex vert s = s
  { graph = case currentVertex s of
      Just old
        | old == vert -> graph s
        | otherwise -> M.insertWith S.union old (S.singleton vert) (graph s)
      Nothing -> graph s
  , currentVertex = Just vert
  }

data NextStep = Restart
              | Continue Bool

handleDecision :: DecisionNumber -> [Bool] -> State -> (NextStep, State)
handleDecision decision trace s =
  let curSystemState = fromJust (currentVertex s) -- FIXME fromJust
      decisionState = M.lookup (curSystemState, decision) (decisions s)
  in case decisionState of
    Nothing -> (Continue True, s { decisions = M.insert (curSystemState, decision) (Half trace) (decisions s) })
    Just (Half _) -> (Continue False, s { decisions = M.insert (curSystemState, decision) Full (decisions s) })
    Just Full -> (Restart, s)

findOpenDecision :: State -> Maybe [Bool]
findOpenDecision = listToMaybe . mapMaybe isHalf . M.elems . decisions
  where
    isHalf (Half trace) = Just trace
    isHalf _            = Nothing



driver :: FilePath -> IO ()
driver prog = driver' prog (Vertex "start" (Metadata "" ""))

driver' :: FilePath -> Vertex -> IO ()
driver' prog startVertex = do
  state <- newIORef (State M.empty (Just startVertex) M.empty [])
  oneProgRun prog state Nothing

  forever $ do
    modifyIORef state $ \s -> s { curTrace = [], currentVertex = Just startVertex }
    findOpenDecision <$> readIORef state >>= \case
      Nothing    -> do
        s <- readIORef state
        T.putStrLn (graph2Dot (graph s))
        exitSuccess
      Just trace ->
        oneProgRun prog state (Just $ reverse trace)

oneProgRun :: FilePath -> IORef State -> Maybe [Bool] -> IO ()
oneProgRun prog state trace = do
  (Just pstdin, Just pstdout, _, handle) <- createProcess (proc prog []) { std_in = CreatePipe, std_out = CreatePipe }
  case trace of
    Just trace' -> tracePath pstdin pstdout state trace'
    _ -> return ()

  let loop = handleLine pstdout state >>= \case
        Decision x -> do
          s <- readIORef state
          let (next, s') = handleDecision x (curTrace s) s
          writeIORef state s'
          case next of
            Restart -> terminateProcess handle >> mapM_ hClose [pstdin, pstdout]
            Continue x' -> do
              makeDecision pstdin state x'
              loop

        EOF -> terminateProcess handle >> mapM_ hClose [pstdin, pstdout]
        NewState _ -> loop

  loop


tracePath :: Handle -> Handle -> IORef State -> [Bool] -> IO ()
tracePath _      _       _     [] = return ()
tracePath pstdin pstdout state (decision:ds) =
  handleLine pstdout state >>= \case
    EOF -> error "premature eof" -- FIXME
    Decision _ -> do
      makeDecision pstdin state decision
      tracePath pstdin pstdout state ds
    NewState _ -> tracePath pstdin pstdout state (decision:ds)

makeDecision :: Handle -> IORef State -> Bool -> IO ()
makeDecision pstdin state decision = do
  hPutStrLn pstdin $ if decision then "1" else "0"
  hFlush pstdin
  modifyIORef state (\s -> s { curTrace = decision : curTrace s})
  hPutStrLn stderr $ "deciding " ++ show decision

data LineOut = NewState Vertex
             | Decision DecisionNumber
             | EOF

parseLine :: Handle -> IO LineOut
parseLine handle = do
  eof <- hIsEOF handle
  if eof
    then return EOF
    else do
      line <- T.hGetLine handle
      T.hPutStrLn stderr $ "Found line: " <> line
      case T.words line of
        ("Decision":num:_) -> return $ Decision (read $ T.unpack num)
        ["Function",func,"before","syscall",call,"in","state",state]
          -> return $ NewState $ Vertex state (Metadata func call)
        _ -> parseLine handle -- ignore unknown output

handleLine :: Handle -> IORef State -> IO LineOut
handleLine pstdin state = parseLine pstdin >>= \case
  NewState vert -> modifyIORef state (addVertex vert) >> return (NewState vert)
  other         -> return other

graph2Dot :: Graph -> Text
graph2Dot graph =
     "digraph gcfg {\n"
  <> dotNodes graph
  <> M.foldMapWithKey nodeStr graph
  <> "}"

  where
    nodeStr from tos = mconcat $ map (arrow from) $ S.toList tos
    arrow from to = "  " <> nodeId from <> " -> "
      <> nodeId to <> " [label=\"" <> edgeLabel (metaSyscall $ vertMeta from) <> "\"];\n"

-- | Generate readable edge label from syscall name
--
-- Transforms OSEKOS_foo_bar into foo(bar)
edgeLabel :: Text -> Text
edgeLabel (T.stripPrefix "OSEKOS_" -> Just txt) =
  let (call, argument) = T.breakOnEnd "_" txt
      derscore = T.dropAround (=='_')  -- un-underscore == derscore
  in derscore call <> "(" <> derscore argument <> ")"
edgeLabel txt = txt -- There are "syscalls" that don't begin with OSEKOS_. E.g ISRs

dotNodes :: Graph -> Text
dotNodes = foldMap formatNode . allNotes
  where
    allNotes m = M.keysSet m `S.union` fold m
    formatNode vert = nodeId vert
                   <> " [label=\""
                   <> "State: " <> vertState vert
                   <> "\\nFunction: " <> stripPrefix' "OSEKOS_" (metaFunc (vertMeta vert))
                   <> "\", shape=box];\n"

nodeId :: Vertex -> Text
nodeId vert = "node" <> vertState vert <> "_" <> metaSyscall (vertMeta vert)

-- | Strip prefix but return original string if it doesn't begin with the prefix
stripPrefix' :: Text -> Text -> Text
stripPrefix' pref (T.stripPrefix pref -> Just txt) = txt
stripPrefix' _ txt = txt
