{-# LANGUAGE OverloadedStrings, QuasiQuotes #-}

module CFG.CSpec (spec) where

import           Data.Monoid
import           Data.Text (Text)
import qualified Data.Text as T
import qualified Data.IntSet as S
import           Test.Hspec
import           Data.String.Here
import           Control.Monad.State

import           CFG.IR
import           CFG.C
import           CFG.Types

spec :: Spec
spec = do
  instrsToCTest
  interruptsSpec
  compileToCTest
  compileToCAllTest
  externDecls

example1Input :: [Instr]
example1Input =
  [ Goto "Two"
  , Label "Zero"
  , Return
  , Label "One"
  , IfThanElse "Two" "Zero"
  , Call "syscall" TypeSyscall False
  , Label "Two"
  , Goto "One"
  ]

example1OutputLines :: Int -- ^ Decision number used for decision
                    -> [Text]
example1OutputLines dec =
  [ "goto Two;"
  , "Zero:"
  , "return;"
  , "One:"
  , "if (_decisionMaker(" <> T.pack (show dec) <> "))"
  , "  goto Two;"
  , "else"
  , "  goto Zero;"
  , "_print_os_state(\"syscall\", __func__);"
  , "syscall();"
  , "Two:"
  , "goto One;"
  ]

example1Output :: Int -- ^ Decision number
               -> Text
example1Output dec = T.unlines (example1OutputLines dec)

example1Function :: Function [Instr]
example1Function = Function
  { funName = "foo"
  , funKind = KindFunction
  , funSubtask = undefined -- irrelevant for this test
  , funEntryNode = undefined -- irrelevant for this test
  , funBlocks = undefined -- irrelevant for this test
  , funAnnotation = example1Input
  }

example1FunctionC :: Int -> Text
example1FunctionC dec =
     "void foo() {\n"
  <> T.unlines (map ("  " <>) (example1OutputLines dec))
  <> "}\n"

example1FunctionCDecl :: Text
example1FunctionCDecl = "void foo();\n"

example1TaskC :: Int -> Text
example1TaskC dec =
     "TASK(foo) {\n"
  <> T.unlines (map ("  " <>) (example1OutputLines dec))
  <> "}\n"

example1TaskCDecl :: Text
example1TaskCDecl = "DeclareTask(foo);\n"

example1ISRC :: Int -> Text
example1ISRC dec =
     "ISR2(foo) {\n"
  <> T.unlines (map ("  " <>) (example1OutputLines dec))
  <> "}\n"

example2Function :: Function [Instr]
example2Function = example1Function { funName = "bar" }

example2FunctionC :: Int -> Text
example2FunctionC dec =
     "void bar() {\n"
  <> T.unlines (map ("  " <>) (example1OutputLines dec))
  <> "}\n"

example2FunctionCDecl :: Text
example2FunctionCDecl = "void bar();\n"

example2TaskC :: Int -> Text
example2TaskC dec =
     "TASK(bar) {\n"
  <> T.unlines (map ("  " <>) (example1OutputLines dec))
  <> "}\n"

example2TaskCDecl :: Text
example2TaskCDecl = "DeclareTask(bar);\n"

example2ISRC :: Int -> Text
example2ISRC dec =
     "ISR2(bar) {\n"
  <> T.unlines (map ("  " <>) (example1OutputLines dec))
  <> "}\n"

exampleExternDecls :: Text
exampleExternDecls = "extern \"C\" {\nextern void syscall();\n}"

-- | Run DecisionM monad with 0 as initial state
runDecision :: DecisionM a -> a
runDecision = flip evalState 0

instrsToCTest :: Spec
instrsToCTest = describe "instrsToC" $ do
  it "works for a given example" $
    runDecision (instrsToC Nothing KindFunction example1Input) `shouldBe` example1Output 0

compileToCTest :: Spec
compileToCTest = describe "compileToC" $ do
  it "works with a function" $
    runDecision (compileToC Nothing example1Function) `shouldBe` example1FunctionC 0

  it "works with a task" $
    runDecision (compileToC Nothing example1Function { funKind = KindSubtask }) `shouldBe` example1TaskC 0

  it "works with an isr" $
    runDecision (compileToC Nothing example1Function { funKind = KindISR }) `shouldBe` example1ISRC 0

include :: Text
include = "#include <mockup_prelude.cc>"

emptyIdleHook :: Text
emptyIdleHook = [here|
void PreIdleHook() {
  _print_os_state("PreIdleHook", __func__);
  ShutdownMachine();
}
|]

compileToCAllTest :: Spec
compileToCAllTest = describe "compileToCAll" $ do
  it "compiles two functions" $
    compileToCAll Nothing [example1Function, example2Function]
      `shouldBe` ( include <> "\n\n" <> exampleExternDecls <> "\n"
                 <> example1FunctionCDecl <> example2FunctionCDecl <> "\n"
                 <> example1FunctionC 0 <> "\n"
                 <> example2FunctionC 1 <> "\n"
                 <> emptyIdleHook <> "\n"
                 )

  it "compiles two tasks" $
    compileToCAll Nothing [example1Function { funKind = KindSubtask },
                   example2Function { funKind = KindSubtask }
                  ]
      `shouldBe` (  include <> "\n\n" <> exampleExternDecls <> "\n"
                 <> example1TaskCDecl <> example2TaskCDecl <> "\n"
                 <> example1TaskC 0 <> "\n"
                 <> example2TaskC 1 <> "\n"
                 <> emptyIdleHook <> "\n"
                 )

  it "compiles one task and one function" $
    compileToCAll Nothing [example1Function { funKind = KindSubtask },
                   example2Function
                  ]
      `shouldBe` (  include <> "\n\n" <> exampleExternDecls <> "\n"
                 <> example1TaskCDecl <> example2FunctionCDecl <> "\n"
                 <> example1TaskC 0 <> "\n"
                 <> example2FunctionC 1 <> "\n"
                 <> emptyIdleHook <> "\n"
                 )

  it "compiles one tasks and one ISR" $ do
    compileToCAll Nothing [example1Function { funKind = KindSubtask },
                   example2Function { funKind = KindISR }
                  ]
      `shouldBe` (  include <> "\n\n" <> exampleExternDecls <> "\n"
                 <> example1TaskCDecl <> "\n"
                 <> example1TaskC 0 <> "\n"
                 <> example2ISRC 1 <> "\n"
                 <> emptyIdleHook <> "\n"
                 )


declFunction1 :: Function [Instr]
declFunction1 = Function
  { funKind = KindFunction
  , funName = "foo"
  , funSubtask = "" -- irrelevant
  , funEntryNode = "" -- irrelevant
  , funBlocks = [] -- irrelevant
  , funAnnotation = [ Call "bar" TypeFunction False ]
  }

declFunction2 :: Function [Instr]
declFunction2 = Function
  { funKind = KindFunction
  , funName = "test"
  , funSubtask = "" -- irrelevant
  , funEntryNode = "" -- irrelevant
  , funBlocks = [] -- irrelevant
  , funAnnotation = [ Call "foo" TypeFunction False ]
  }

externDecls :: Spec
externDecls = describe "declareExtern" $ do
  it "declares a called function" $
    let output = [here|
extern "C" {
extern void bar();
}
|]
    in declareExtern [declFunction1] `shouldBe` output

  it "doesn't declare a defined function" $
    let output = [here|
extern "C" {
extern void bar();
}
|]
    in declareExtern [declFunction1, declFunction2] `shouldBe` output


interruptsSpec :: Spec
interruptsSpec = do
  describe "instrsToC" $
    context "with interrupts" $ do

      it "doesn't trigger interrupts before kickoff" $
        runDecision (instrsToC (Just $ S.fromList [42]) KindFunction [Call "foobar" TypeSyscall True])
        `shouldNotContain'` "interrupt"

      it "doesn't trigger interrupts in ISRs" $
        runDecision (instrsToC (Just $ S.fromList [23]) KindISR [Call "foo" TypeSyscall False])
        `shouldNotContain'` "interrupt"

      it "triggers all available interrupts" $
        runDecision (instrsToC (Just $ S.fromList [23, 42]) KindSubtask [Call "foo" TypeSyscall False])
        `shouldSatisfy` (\x -> "23" `T.isInfixOf` x && "42" `T.isInfixOf` x)

  describe "preIdleHook" $
    it "triggers all interrupts" $
      runDecision (preIdleHook (Just $ S.fromList [23, 42]))
      `shouldSatisfy` (\x -> "23" `T.isInfixOf` x && "42" `T.isInfixOf` x)


-- | Helper function like shouldNotContain from HSpec but with Text instead of String
shouldNotContain' :: Text -> Text -> Expectation
shouldNotContain' heystack needle = heystack `shouldNotSatisfy` (needle `T.isInfixOf`)
