{-# LANGUAGE OverloadedStrings, DuplicateRecordFields, QuasiQuotes #-}
module CFG.ParserSpec (spec) where

import           Data.Aeson as J
import           Data.Monoid
import           Data.Text (Text)
import qualified Data.Text as T
import qualified Data.IntSet as S
import           Test.Hspec
import qualified Data.ByteString.Lazy as BS
import           Data.String.Here

import           CFG.Parser
import           CFG.Types

spec :: Spec
spec = do
  example1
  example2
  both
  isrExample

example1Input :: BS.ByteString
example1Input = [here|
{ "name": "Handler12",
  "kind": "subtask",
  "subtask": "Handler12",
  "entry": "ABB40/kickoff",
  "abbs": [ { "name": "ABB37/TerminateTask",
	"call": {"type": "syscall", "name": "OSEKOS_TerminateTask_BB134"},
	"isKickoff": false,
	"successors": ["ABB38"]
      },
      { "name": "ABB38",
	"call": null,
	"isKickoff": false,
	"successors": []
      },
      { "name": "ABB39",
	"call": null,
	"isKickoff": false,
	"successors": ["ABB37/TerminateTask"]
      },
      { "name": "ABB40/kickoff",
	"call": {"type": "syscall", "name": "OSEKOS_kickoff_1"},
	"isKickoff": true,
	"successors": ["ABB39"]
      }
    ]
}
|]

example1Output :: [Function ()]
example1Output = [
  Function
    { funKind = KindSubtask
    , funName = "Handler12"
    , funSubtask = "Handler12"
    , funEntryNode = "ABB40/kickoff"
    , funBlocks =
      [ Block "ABB37/TerminateTask" ["ABB38"] (Just $ SystemCall "OSEKOS_TerminateTask_BB134") False
      , Block "ABB38" [] Nothing False
      , Block "ABB39" ["ABB37/TerminateTask"] Nothing False
      , Block "ABB40/kickoff" ["ABB39"] (Just $ SystemCall "OSEKOS_kickoff_1") True
      ]
    , funAnnotation = ()
    }]

example1 :: Spec
example1 = it "parses a function" $ do
  let json = "{\"interrupts\": [], \"functions\": [" <> example1Input <> "]}"
  parseFile json `shouldBe` Right (Application example1Output S.empty)

example2Input :: BS.ByteString
example2Input = [here|
{ "name": "Handler11",
  "kind": "subtask",
  "subtask": "Handler11",
  "entry": "ABB24/kickoff",
  "abbs": [ { "name": "ABB9/ActivateTask",
	"call": {"type": "syscall", "name": "OSEKOS_ActivateTask_BB118"},
	"isKickoff": false,
	"successors": ["ABB10"]
      },
      { "name": "ABB10",
	"call": {"type": "function", "name": "_Z14print_os_statei"},
	"isKickoff": false,
	"successors": ["ABB31", "ABB32"]
      },
      { "name": "ABB15/ActivateTask",
	"call": {"type": "syscall", "name": "OSEKOS_ActivateTask_BB124"},
	"isKickoff": false,
	"successors": ["ABB16"]
      },
      { "name": "ABB16",
	"call": null,
	"isKickoff": false,
	"successors": ["ABB33"]
      },
      { "name": "ABB17",
	"call": {"type": "function", "name": "_Z12foo_functionv"},
	"isKickoff": false,
	"successors": ["ABB18"]
      },
      { "name": "ABB18",
	"call": null,
	"isKickoff": false,
	"successors": ["ABB33"]
      },
      { "name": "ABB21/TerminateTask",
	"call": {"type": "syscall", "name": "OSEKOS_TerminateTask_BB130"},
	"isKickoff": false,
	"successors": ["ABB22"]
      },
      { "name": "ABB22",
	"call": null,
	"isKickoff": false,
	"successors": []
      },
      { "name": "ABB23",
	"call": null,
	"isKickoff": false,
	"successors": ["ABB9/ActivateTask"]
      },
      { "name": "ABB24/kickoff",
	"call": {"type": "syscall", "name": "OSEKOS_kickoff_0"},
	"isKickoff": true,
	"successors": ["ABB23"]
      },
      { "name": "ABB31",
	"call": null,
	"isKickoff": false,
	"successors": ["ABB15/ActivateTask"]
      },
      { "name": "ABB32",
	"call": null,
	"isKickoff": false,
	"successors": ["ABB17"]
      },
      { "name": "ABB33",
	"call": null,
	"isKickoff": false,
	"successors": ["ABB21/TerminateTask"]
      }
    ]
}
|]

example2Output :: [Function ()]
example2Output = [
  Function
    { funKind = KindSubtask
    , funName = "Handler11"
    , funSubtask = "Handler11"
    , funEntryNode = "ABB24/kickoff"
    , funBlocks =
      [ Block "ABB9/ActivateTask" ["ABB10"] (Just $ SystemCall "OSEKOS_ActivateTask_BB118") False
      , Block "ABB10" ["ABB31", "ABB32"] (Just $ FunctionCall "_Z14print_os_statei") False
      , Block "ABB15/ActivateTask" ["ABB16"] (Just $ SystemCall "OSEKOS_ActivateTask_BB124") False
      , Block "ABB16" ["ABB33"] Nothing False
      , Block "ABB17" ["ABB18"] (Just $ FunctionCall "_Z12foo_functionv") False
      , Block "ABB18" ["ABB33"] Nothing False
      , Block "ABB21/TerminateTask" ["ABB22"] (Just $ SystemCall "OSEKOS_TerminateTask_BB130") False
      , Block "ABB22" [] Nothing False
      , Block "ABB23" ["ABB9/ActivateTask"] Nothing False
      , Block "ABB24/kickoff" ["ABB23"] (Just $ SystemCall "OSEKOS_kickoff_0") True
      , Block "ABB31" ["ABB15/ActivateTask"] Nothing False
      , Block "ABB32" ["ABB17"] Nothing False
      , Block "ABB33" ["ABB21/TerminateTask"] Nothing False
      ]
    , funAnnotation = ()
    }]

example2 :: Spec
example2 = it "parses another function" $ do
  let json = "{\"interrupts\": [], \"functions\": [" <> example2Input <> "]}"
  parseFile json `shouldBe` Right (Application example2Output S.empty)


both :: Spec
both = it "parses both functions" $ do
  let json = "{\"interrupts\": [], \"functions\": ["
             <> example1Input <> ", " <> example2Input <> "]}"
  parseFile json `shouldBe` Right (Application (example1Output <> example2Output) S.empty)

isrExampleInput :: BS.ByteString
isrExampleInput = [here|
{
  "kind": "isr",
  "entry": "ABB197/kickoff",
  "abbs": [
    {
      "isKickoff": false,
      "call": null,
      "successors": [
        "ABB5/ActivateTask"
      ],
      "name": "ABB4"
    },
    {
      "isKickoff": false,
      "call": {
        "type": "syscall",
        "name": "OSEKOS_ActivateTask_BB35"
      },
      "successors": [
        "ABB6"
      ],
      "name": "ABB5/ActivateTask"
    },
    {
      "isKickoff": false,
      "call": null,
      "successors": [
        "ABB195/iret"
      ],
      "name": "ABB6"
    },
    {
      "isKickoff": false,
      "call": {
        "type": "syscall",
        "name": "OSEKOS_iret__ABB195"
      },
      "successors": [],
      "name": "ABB195/iret"
    },
    {
      "isKickoff": true,
      "call": {
        "type": "syscall",
        "name": "OSEKOS_kickoff__ABB197"
      },
      "successors": [
        "ABB4"
      ],
      "name": "ABB197/kickoff"
    }
  ],
  "subtask": "ISR1",
  "name": "ISR1"
}
|]

isrExampleOutput :: [Function ()]
isrExampleOutput = [
  Function
    { funKind = KindISR
    , funName = "ISR1"
    , funSubtask = "ISR1"
    , funEntryNode = "ABB197/kickoff"
    , funBlocks =
      [ Block "ABB4" ["ABB5/ActivateTask"] Nothing False
      , Block "ABB5/ActivateTask" ["ABB6"] (Just $ SystemCall "OSEKOS_ActivateTask_BB35") False
      , Block "ABB6" ["ABB195/iret"] Nothing False
      , Block "ABB195/iret" [] (Just $ SystemCall "OSEKOS_iret__ABB195") False
      , Block "ABB197/kickoff" ["ABB4"] (Just $ SystemCall "OSEKOS_kickoff__ABB197") True
      ]
    , funAnnotation = ()
    }
  ]

isrExample :: Spec
isrExample = it "parses an ISR" $ do
  let json = "{\"interrupts\": [4, 5], \"functions\": [" <> isrExampleInput <> "]}"
  parseFile json `shouldBe` Right (Application isrExampleOutput (S.fromList [4, 5]))
